/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the Open-Xchange, Inc. group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2004-2014 Open-Xchange, Inc.
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.guard.server;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.gson.Gson;
import com.google.gson.JsonObject;
import com.openexchange.guard.database.Access;
import com.openexchange.guard.encr.EncrLib;


public class UserData {

	private static Logger logger = LoggerFactory.getLogger(UserData.class);
	
    public int userid;

    public int cid;

    public String ox_password;

    public String encr_password;

    public UserData() {

    }

    public UserData(String auth, com.openexchange.guard.server.OxCookie cookie) {
        long start = System.currentTimeMillis();
        userid = -1;
        ox_password = encr_password = null;
        try {
            Access ac = new Access();
            String token = ac.getToken(cookie.JSESSIONID);
            EncrLib encr = new EncrLib();
            if (token == null)
                return;
            String userdata = encr.decryptAES(auth, token);
            if (userdata.equals("")) {
                return;// Bad authentication
            }
            Gson gson = new Gson();
            JsonObject userjson = gson.fromJson(userdata, JsonObject.class);
            userid = userjson.get("user_id").getAsInt();
            cid = userjson.get("cid").getAsInt();
            if (userjson.get("ox_password") != null) {
                ox_password = userjson.get("ox_password").getAsString();
            }
            encr_password = userjson.get("encr_password").getAsString();
        } catch (Exception e) {
            logger.error("Error creating user data for " + userid, e);
        }
        logger.info("Decode user data at " + start);
    }
}
